TITLE receiver

; YABRI - Yet Another Bridge for Ethernet
;
; Receiver.asm V0.3p1
; (c) Carsten Gross, Rainer Bawidamann 1995
;
; History
;
; 29.6.95
;    changed from handle to driver number
;    export lookup, enhanced output
; 7.7.95
;    some bugfixes
; 15.7.95
;    new repeater mode
; 5.8.95
;    new variable _adr_cnt
; 21.9.95
;    switched to english (*uff*)

Ideal
MODEL SMALL
p286

puffers   equ   020      ; number of receive buffers
Timeout   equ   5*60*19  ; 5*60 sec, 18.2 IRQs per sec


segment   _DATA

	   db 'Carsten Gross & Rainer Bawidamann 1995'
Zaehler    dw 0
public _screen,_Ethernetad
_screen     dw  0b000h   ; address of screen memory
_Ethernetad dw  2020h,2020h,2020h

IntrpCS    dw ?          ; pointer of the int routine
IntrpIP    dw ?
TimeOutCount dw ?        ; counter for timeout

; receive buffers are now declared in the C part
; still here as comment to show buffer layout
public _recvbuf
_recvbuf  dw (1520/2)*puffers dup(0)
	  ; receive buffer structure:
	  ; 1514 data bytes
	  ; flag (what to do with this packet)
	  ; number of the driver
	  ; length of the packet in the buffer

; lookup table for 255 ethernet addresses
public _lookup,_end_lookup
_lookup    db 255*10 dup (0)
	  ; 6 Bytes ethernet address
	  ; 2 Bytes flag and driver number
	  ; 2 Bytes timeout (for future use)
_end_lookup dw _lookup

public _adr_cnt
_adr_cnt    dw 0

ends      _DATA

extrn _fatal:word
;extrn _recvbuf:word
extrn _repeater:word

segment   _TEXT

; entries for packet drivers when they received a packet
; one entry per driver to distinguish

public _receiver1, _receiver2, _receiver3, _receiver4, _receiver5
proc _receiver1    far
	  mov bx,1
	  jmp receiver
endp

proc _receiver2    far
	  mov bx,2
	  jmp receiver
endp

proc _receiver3    far
	  mov bx,3
	  jmp receiver
endp

proc _receiver4    far
	  mov bx,4
	  jmp receiver
endp

proc _receiver5    far
	  mov bx,5
	  jmp receiver
endp


; for the format of the output of the asm part see yabri.c

proc   receiver        far
       assume ds:_DATA,cs:_TEXT
	  pushf
	  cli
	  push ds
	  cmp ax,0      ;0 -> driver wants a buffer for the packet
	  jz getbuf     ;search a free buffer
	  cmp ax,1      ;1-> driver copied packet into buffer
	  jz dispatch   ;look what to do with the packet
error:
	  xor di,di
	  mov es,di     ;es:di == 0 -> error report to the driver
	  pop ds
	  popf
	  ret

; search a free buffer and return address to driver

getbuf:
	  cmp cx,1514   ;can we handle the packet ?
	  jg error      ;no, too big

	  push bx       ;save driver number
	  mov bx,_DATA
	  mov ds,bx     ;set data segment
	  xor bx,bx     ;bx holds now relative pointer to buffer
	  xor cx,cx     ;the number of the current buffer
l1:
	  ;is the buffer bx/cx free ?
	  cmp [word ptr (offset _recvbuf)+bx+1514+2],0
	  je Puf_gefunden   ;yes
	  add bx,1520       ;no, check next buffer
	  inc cx
	  cmp cx,puffers    ;last buffer tested ?
	  jl l1
	  pop bx
	  mov cl,'l'        ;yes, show 'packet lost'
	  call showpacket
	  jmp error         ;report it error to driver

Puf_gefunden:   ;buffer found, reserve it and tell driver the address

	  mov si,bx         ;we can't we use every register for every instruction???
	  pop bx            ;remember driver number
	  ;set driver number. this reserves the buffer
	  mov [offset _recvbuf+1514+2+si],bx
	  push ds           ;ds is the first part of the address
	  pop es            ;return address in es:di
	  mov di,offset _recvbuf
	  add di,si
	  pop ds
	  popf
	  ret

; process packet in buffer

	  assume ds:_DATA,cs:_TEXT
dispatch:
	  cmp bx,[si+1514+2]    ;buffer reserved for this driver?
          je @@l30
	  jmp fatal_err         ;no -> error (in yabri?)
@@l30:
	  mov [si+1514+4],cx    ;save packet length

; ok, packet is in buffer now. now look what we want to do with the packet
; we distinguish the following cases:
;
; broadcast   the packet has the ethernet broadcast address (ff:ff:ff:ff:ff:ff)
;             normally we send it to every other driver
;             mutlicast is treated like broadcast
; ARP         was treated differently in previous versions. now we just
;             show the occurence with a 'a' in the top line
; others      1. if the destination address is in the lookup table we get
;               the line (driver) the machine is on from there
;               if it is the line the packet was received drop the packet,
;               else send the packet the corresponding line. as sending takes
;               some time we do this in the C part. we store the necessary
;               information in the flag field (named TxFlag in C)
;             2. if the address is not in the table we send it to every
;               other line just like we do with broadcast packets
; in general, we insert the source address of every packet in out lookup
; table, as we know now to which line this machine is connected (it's the
; line the packet came from)


; first test: are we in repeater mode?
	  cmp [word ptr _repeater],0
	  je no_repeat

; yes, show the packet and send it to every other line
	  call Check_Ether	;make usual tests even in repeater mode.
	  jnc @@l20             ;Is it `our' Ethernetaddress?
	  or bx,4000h           ;set flag for the C part
	  mov cl,'!'            ;tell user about it
	  jmp @@l21
@@l20:
	  mov cl,'.'
@@l21:
	  call showpacket
	  jmp broadcast         ;and finally send like a broadcast

; is this packet a ARP packet?
; this information is in the byte in position 12

no_repeat:
	  cmp [word ptr si+12],0608h         ;0608==Arp (network order)
	  jne Kein_ARP
	  mov cl,'a'                         ;show it
	  call showpacket
Kein_ARP:
	  call Eintrag                       ;we try to insert every packet

; second test: is this packet broadcast?
	  call Check_Ether      ;test if this is our ethernet address
	  jnc @@l3              ;no ->
	  or bx,4000h           ;set flag for the C part
	  mov cl,'!'            ;tell user about it
	  call showpacket
	  jmp normale_ausgabe   ;rest like 'normal' packets

@@l3:
	  mov ax,[si]           ;LSB of 1. byte == 1 -> broadcast/multicast
	  and al,00000001b
	  je kein_broadcast     ;not multicast/broadcast, transmit normally
	  mov cl,'b'            ;show broadcast packet
	  call showpacket

; set flag for broadcast and let the C part do the rest
broadcast:
	  or bx,8000h           ;set broadcast flag

; the exit if packet is scheduled to the C part
normale_ausgabe:
	  mov [si+1514],bx      ;save line number
	  pop ds
	  popf
	  ret

; the 'normal' case: search line for the packet
kein_broadcast:
	  call durchsuchen    ;search in lookup table
	  jc no_broadcast     ;found ->
	  mov cl,'u'          ;show unknown packet
	  call showpacket
	  jmp broadcast       ;treat like broadcast

no_broadcast:
	  cmp bx,[si+1514+2]  ;same net as packet came from?
	  je drop_it          ;yes, forget packet
	  mov cl,'s'          ;show a normal packet
	  call showpacket
	  jmp normale_ausgabe ;and exit

; we don't need to send this packet, just free the buffer
drop_it:
	  mov cl,'r'                     ;show action
	  call showpacket
	  mov [word ptr si+1514+4],0     ;set length to 0
	  mov [word ptr si+1514+2],0     ;clear line number -> buffer free
	  pop ds
	  popf
	  ret

; we get here in case of an error. tell C part about it!
fatal_err:
	  mov [word ptr _fatal],0fffeh   ;set variale fatal
	  pop ds
	  popf
	  ret
endp


; insert the source address if not yet in the table
; input:  ds:si         pointer to packet buffer
;         bx            line/driver number
;         [_lookup]     start of lookup table
;         [_end_lookup] end of table
; changing ax,dx !

proc      Eintrag      near
	  assume ds:_DATA,cs:_TEXT
	  mov dx,bx             ;saveline number
	  push si               ;save buffer pointer
	  add si,6              ;set pointer to source address
	  call durchsuchen      ;search source address in table
	  pop si                ;restore address
	  jc @@l2               ;already in table ->

; now we want to insert the address. first we check if there is place for it
	  mov bx,[offset _end_lookup]    ;the actual end of the table
	  cmp bx,offset _lookup+(10*254) ;reached maximum?
	  jl  @@l3                       ;no ->

; table is full, clear it (ok, this is not the best solution. do YOU like
; to program x86 asm?
	  call flush_table

@@l3:
	  mov ax,[si+6]         ;copy address to first free entry
	  mov [bx],ax
	  mov ax,[si+8]
	  mov [bx+2],ax
	  mov ax,[si+10]
	  mov [bx+4],ax
	  mov [bx+6],dx         ;set line for this entry
	  mov [word ptr bx+8],0ffffh  ;preset timeout counter (not used)
	  add bx,10             ;set new table end
	  mov [ds:offset _end_lookup],bx
; set the number of entries for the C part
	  inc [ds:(word ptr (offset _adr_cnt))]
@@l2:
	  mov bx,dx             ;restore line number
	  ret
endp


; search an ethernet in lookup table
; input:  ds:si         pointer to address
;         [_lookup]     the lookup table
;         [_end_lookup] end of the table
; output: carry=0       address found
;              =1       not found
;         bx            line/driver number
; changes ax; changes bx only if address found


proc      durchsuchen            near
	  push di
	  mov di,offset _lookup ;di is the pointer to search in the table
@@l1:
	  mov ax,[di+4]         ;compare address with entry
	  cmp [si+4],ax
	  jne @@l3
	  mov ax,[di+2]
	  cmp [si+2],ax
	  jne @@l3
	  mov ax,[di]
	  cmp [si],ax
	  je found              ;is equal, return line number ->
@@l3:
	  add di,10             ;check next entry
	  cmp di,[ds:offset _end_lookup]
	  jle @@l1              ;not end -> contiune
	  pop di
	  clc                   ;set flag 'not found'
	  ret
found:
	  mov bx,[di+6]         ;get line number form table
	  pop di
	  stc                   ;set flag 'found'
	  ret
endp

; tell the user that we have a packet and what we do with it
; input: cl  character to print
;        bx  line number to print
; changes ax

proc      showpacket    near
	  assume es:nothing,ds:_DATA,cs:_TEXT
	  push es
	  push bx
	  mov ax,[(offset _screen)]             ;address of the screen
	  mov es,ax
	  cmp [Word Ptr offset Zaehler],0160d*3 ;end of print area?
	  jl @@l1
	  mov [Word Ptr offset Zaehler],0       ;yes, continue at 0,0
@@l1:
	  mov ax,bx                             ;first print line number
	  mov bx,[Word Ptr offset Zaehler]      ;the actual position
	  add ax,0730h                          ;change to appropriate char
	  mov [es:bx],ax                        ;to screen
	  mov [es:bx+2],cl                      ;then the action character
	  mov [es:(Word Ptr (bx+4))],0720h      ;and a space
	  add bx,4                              ;inc position (2 chars)
	  mov [Word Ptr offset Zaehler],bx
	  pop bx
	  pop es
	  ret
endp

; clear lookup table, entry for C part
public _flush_table
proc _flush_table far
	  assume cs:_TEXT,ds:_DATA
	  push ds                  ;need to preserve ds!
	  mov ax,_DATA
	  mov ds,ax
	  cli                      ;flush_table shouldn't be interrupted
	  call flush_table
	  sti
	  pop ds
	  ret
endp

; clear the lookup table
; changes ax,cl
proc      flush_table   near
	  assume cs:_TEXT,ds:_DATA
	  push bx
	  push cx
	  mov [ds:(word ptr (offset _end_lookup))],(offset _lookup)
	  mov [ds:(word ptr (offset _adr_cnt))],0
	  mov bx,0
	  mov cl,'f'
	  call showpacket
@@l1:
; clear the first entry
	  mov [ds:(Word ptr (offset _lookup)+bx)],0
	  inc bx
	  inc bx
	  cmp bx,8
	  jle @@l1
; preset the timeout (not used)
	  mov [ds:(Word Ptr (offset _lookup)+8)],0ffffh
	  pop cx
	  pop bx
	  ret
endp


; the routines for the timer (table_flush every 5 min)

public _get_TimerIRQ,_release_TimerIRQ

; install an int handler, called from C as 'get_TimerIRQ()'
proc     _get_TimerIRQ  far
	  assume cs:_TEXT,ds:_DATA
masm
	  push ds
	  mov ax,_DATA
	  mov ds,ax
	  mov [Word Ptr (offset TimeOutCount)],Timeout
	  cli
	  mov ax,0
	  mov es,ax
	  mov ax,es:[070h]              ;70h is the pointer of INT 01ch
	  mov [IntrpIP],ax              ;save old pointer
	  mov ax,es:[072h]
	  mov [IntrpCS],ax
	  mov ax,offset Timeout_IRQ
	  mov es:[070h],ax              ;and set own
	  mov ax,cs
	  mov es:[072h],ax
	  sti
	  pop ds
	  ret
Ideal
endp

; int routine: clear the lookup table after about 5 min.
proc      Timeout_IRQ   far
	  assume cs:_TEXT,ds:_DATA
	  push ds
	  push ax
	  mov ax,_DATA
	  mov ds,ax
	  dec [Word Ptr offset TimeOutCount]           ;5 min. over?
	  cmp [Word Ptr (offset TimeOutCount)],0
	  jne Ende_IRQ                                 ;no ->
	  call flush_table                             ;yes -> flush
	  mov [Word Ptr (offset TimeOutCount)],Timeout ;and reset timer
Ende_IRQ:
	  pop ax
	  pop ds
	  iret
endp

; de-install timer handler, called from C with 'release_TimerIRQ()'
proc      _release_TimerIRQ     far
	  assume ds:_DATA,cs:_TEXT
	  push ds
	  mov ax,_DATA
	  mov ds,ax
	  mov ax,0
	  mov es,ax
	  cli
masm
	  mov ax,[IntrpIP]          ;set old int pointer
	  mov es:[070h],ax
	  mov ax,[IntrpCS]
	  mov es:[072h],ax
	  sti
	  pop ds
	  ret
Ideal
endp

; check if the destination address of the packet is our address
; output: carry=0  not equal
;              =1  it is our address

proc      Check_Ether   near
	  assume ds:_DATA,cs:_TEXT
	  push ax
	  mov ax,[si+4]
	  cmp ax,[offset _Ethernetad+4]
	  jne @@l1
	  mov ax,[si+2]
	  cmp ax,[offset _Ethernetad+2]
	  jne @@l1
	  mov ax,[si]
	  cmp ax,[offset _Ethernetad]
	  jne @@l1
	  stc
	  pop ax
	  ret
@@l1:
	  clc
	  pop ax
	  ret
endp      ;Check_Ether


ends      _TEXT
END
