/* YABRI - Yet Another Bridge for Ethernet
 * (c) Rainer Bawidamann, Carsten Gross 1995
 * This is free software, you can modify, distribute and copy it under
 * the terms of the GNU General Public License, see COPYING for details
 */

/* The packet driver interface is....
 * public domain by Russell Nelson, nelson@crynwr.com.  Politeness dictates
 * that you leave this notice intact */

char *yabri="Rainer Bawidamann & Carsten Gross 1995";
char *version="v0.3p6 4.5.97";

/* yabri.c
   Version 0.3 */

/* History
 *
 *
 * 6.5.97
 *     Added CPP-Macro "PRODUCTION" for a high-performance yabri -
 *     note: NO ERROR handling if this is enabled! reply.c
 *     enhanced and bugfixed.  (Carsten)
 *
 * 29.6.95
 *    changed from handle to driver number
 *    improved output
 *    for testing: flush_table, TimerIRQ, analyzing of every packet
 *
 * 7.7.95
 *    bug fixes in receiver.asm
 *    some minor corrections in yabri.c
 *
 * 9.7.95
 *    version 0.2:
 *    1. try of ping-request/reply
 *
 * 15.7.95
 *    checksum for the IP Header
 *    new 'repeater-mode' transferring ALL packets
 *
 * 16.7.95
 *    statistic-info added
 *
 * 22.7.95
 *    Oops - IP-Header is in network order
 *    ping works now
 *    changed output/keyboard handling
 * 5.8.95
 *    version 0.3:
 *    ARP reply for ping added
 * 2.9.95
 *    check if packet driver accepts promiscous mode
 *    changed warning if only one packet driver
 *    additional parameters: /help, /?, /timer
 *
 * 21.9.95
 *    switch to english completly
 *    new sprint_IP routine
 *
 * 11.5.96
 *    added tftp-server to get information via net
 * 
 * 13.5.96
 *   split file into two pices: yabri.c and reply.c
 *   yabri.c holds bridging code and reply.c the action part
 *   (tftpd, ICMP-Messages, arp reply, tcp-connection reset (NOT YET done)) 
 *
 * 9.6.96
 *   added remote control of repeater/bridge mode (put repeater / put bridge)
 *   Real tftp error messages (strings), not only error codes.
 *   New arrangement of GNU statement
 *
 */

/* Comment out the next line while you're installing
 * otherwise no debuging output will be shown.
 * This is for running systems without screen and keyboard */

#undef PRODUCTION

#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <stdlib.h>
#include "pktdrvr.h"
#include "yabri.h"


#ifdef PRODUCTION
#undef debug_ping
#undef debug_arp
#undef DEBUG_UDP
#define HIGHP      /* High Perfomance Yabri if this is defined */
#else /* ifdef PRODUCTION */
#define debug_ping
#define debug_arp
#define DEBUG_UDP
#undef HIGHP
#endif /* PRODUCTION */




struct packetdrv drivers[5];            /* 5 packet drivers */
int drvnum;                             /* 5 drivers are allowed */
unsigned int fatal=0;
unsigned int repeater=0;                /* !=0 for repeating, not bridging */
extern struct pktbuffer recvbuf[PUFFERS];      /* the RxBuffers */
int Timer_on=0;                         /* 1=TimerIRQ on */
struct ip_adr my_ip;                    /* our (bridge) IP address */
int do_arp=0;                           /* answer to ARP reply */
int tftp_port = 0x3fae;
unsigned long ping_cnt=0;               /* amount of ping replys */
unsigned long arp_cnt=0;                /* amount of  ARP replys */
unsigned long tftp_cnt=0;               /* amount of tftp answers */


char *help="\
Output of the first three lines, number=driver, character:\n\
r - packet received, not sent	          b - broadcast packet\n\
s - packet received and sent              u - unknown dest. -> broadcast\n\
l - packet dropped due to full buffer     f - lookuptable flushed\n\
a - ARP packet received                   ! - packet for 'us' received\n\
\n\
input keys:\n\
q - quit yabri                            h - show this help\n\
p - show 'routing table'                  f - flush lookuptable\n\
s - show statistics                       i - info about version and mode\n\
t - switch timer driven lookuptable flush on/off\n\
a - ARP-reply on/off\n\
";

char *strend(char *str)
{
     for(;*str;str++);
     return str;
}

void help_start(char *progname)
{
     printf("%s [/h] [/timer=x] [ip your.ip.number.here]\n",progname);
     printf("\
   /h        shows this help\n\
   /timer=x  flush lookuptable timer driven yes (x=1) / no (x=0)\n\
   ip x.x.x.x   answer to 'ping' for IP-address x.x.x.x\n\
   \n\
   /c       displays copyright notice\n");
}

void copyright_start(char *progname)
{
   printf("%s Yet another bridge for ethernet\n",progname);
   printf("\
   Copyright (C) 1995,96 Rainer Bawidamann, Carsten Gro\n\
   \n\
   This program is free software; you can redistribute it and/or modify\n\
   it under the terms of the GNU General Public License as published by\n\
   the Free Software Foundation; either version 2 of the License, or\n\
   (at your option) any later version.\n\
\n\
   This program is distributed in the hope that it will be useful,\n\
   but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
   GNU General Public License for more details.\n\
\n\
   You should have received a copy of the GNU General Public License\n\
   along with this program; if not, write to the Free Software\n\
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\
\n\
   The authors are reachable via e-mail:\n\
   widi@sol.wohnheim.uni-ulm.de (Rainer Bawidamann)\n\
   carsten@sol.wohnheim.uni-ulm.de (Carsten Gro)\n\
\n");
}

void print_help(void)
{
        printf("%s",help);
}

void printaddr(char *buf, char *eaddr)
{
	static char hex[] = "0123456789abcdef";
        int l;

        for (l=0; l<5; l++) {
            *buf++ = (char) hex[(eaddr[l] & 0xf0) >> 4];
            *buf++ = (char) hex[eaddr[l] & 0xf];
            *buf++ = ':';
        }
        *buf++ = (char) hex[(eaddr[5] & 0xf0) >> 4];
	*buf++ = (char) hex[eaddr[5] & 0xf];
	*buf++ = 0;
}

/* return a string with the IP address from *ip */

char *sprint_IP(struct ip_adr ip)
{
    static char adr[20];

    sprintf(adr,"%u.%u.%u.%u", ip.adr[0], ip.adr[1],
                               ip.adr[2], ip.adr[3]);
    return adr;
}

/* print list of used packet drivers */

void print_drivers(char *buf)
{
    int n;

    *buf = 0;
    for (n=0;n<drvnum;n++) {
        sprintf(strend(buf), "%d. packet driver '%s' at 0x%.2x,  Adr ",
                      n+1,drivers[n].name, drivers[n].intno);
        printaddr(strend(buf), drivers[n].myeaddr);
	strcat(buf, "\n");
    }
    strcat(buf, "\n");
}

/* print info about yabri */

void print_info(char *buf)
{
    int n, cnt=0;

    buf+=sprintf(buf, "yabri version %s\n\n(c) %s\n\n",version,yabri);
#ifdef HIGHP
    sprintf(strend(buf), "Yabri compiled for HIGH-PERFORMANCE\n");
    sprintf(strend(buf), "without keyboard or error handling.\n");
#else
    sprintf(strend(buf), "Yabri compiled in normal mode..\n");
    sprintf(strend(buf), "including debug and keyboard code\n");
#endif
    for (n=0; n<PUFFERS; n++)
	 if (recvbuf[n].drvr) cnt++;
    sprintf(strend(buf), "%u added packet driver, %u/%u packet buffers\n",
	    drvnum, cnt, PUFFERS);
    for (n=0;lookup[n].drvr;n++);
    sprintf(strend(buf), "%u addresses in address table (max. %u)\n",adr_cnt,TAB_ENTRIES);
    sprintf(strend(buf), "\nMode:\n");
    if (do_arp) {
	sprintf(strend(buf), "- sending ARP reply on %s\n",sprint_IP(my_ip) );
    }
    /* if (ping_on) */
    sprintf(strend(buf), "- sending ping reply on ");
    printaddr(strend(buf), Ethernetad);
    sprintf(strend(buf), "\n");

    if (repeater)
       sprintf(strend(buf), "- repeater mode\n");
    if (Timer_on)
       sprintf(strend(buf), "- timer on\n");
    else
       sprintf(strend(buf), "- timer off\n");
    strcat(buf, "\n");
}

/* print statistics of packet drivers */

void print_stat(char *buf)
{
    int n;
    struct statistics far *stat;

    sprintf(buf, "statistics of packet driver(s):\n");
    sprintf(strend(buf), "      Pkt.rx   Pkt.tx   Bytes.rx   Bytes.tx  Err.rx  Err.tx Pkt.lost\n");
    for (n=0; n<drvnum; n++) {
      if (!get_statistics(drivers[n].intno,drivers[n].handle, &stat)) {
	 sprintf(strend(buf), "%d: %9lu%9lu%11lu%11lu%8lu%8lu%9lu\n",n+1,
	       stat->packets_in, stat->packets_out,
	       stat->bytes_in, stat->bytes_out,
	       stat->errors_in, stat->errors_out, stat->packets_lost);
      }
    }
    sprintf(strend(buf), "\n IP statistics: %lu ARP request, %lu Ping request, %lu TFTP requests answered.\n",
	    arp_cnt,ping_cnt, tftp_cnt);
}

/* print address table */

void print_table(char *buf)
{
    int n;

    *buf=0;
    for (n=0;n<adr_cnt;) {
	 printaddr(strend(buf), lookup[n].addr);
	 sprintf(strend(buf), " / %d   ",lookup[n].drvr);
         ++n;
         if (n % 3 == 0) strcat(buf, "\n");
    }
    strcat(buf, "\n");
}

/* search packet driver at int 'intno', initalize it */
/* returns:   0 == ok */
/*            1 == too many packet drivers */
/*            2 == wrong handle (0) */
/*            3 == driver cannot do promiscous mode */

int newpkt(int intno, void far (*receiver)(void))
{
        int n;
        int i,ii,t;
        int version,class,type,number,basic;
        char tmpstr[100];

        for (n=0;n<5;n++)
            if (drivers[n].intno==0) break;
        if (n==5) {
            fprintf(stderr, "too many drivers (max 5)\n");
            return (1);
        }
        drivers[n].intno=intno;

	drivers[n].handle = access_type(intno,
		CL_ETHERNET,		/* has to be an Ethernet driver */
		0xffff,			/* we don't care whose it is. */
		0,			/* we want the first piece of hardware */
		NULL,			/* doesn't matter because we want all */
		0,			/* zero type length, that is, all. */
		receiver);		/* -> our upcall */
        if (drivers[n].handle==0) {
                drivers[n].intno=0;
                fprintf(stderr, "wrong handle <0>\n");
                return (2);
        }
	/* get the adaptor's Ethernet address */
	get_address (intno,drivers[n].handle,drivers[n].myeaddr,
                     sizeof drivers[n].myeaddr);
        /* check if there's another board with the same ethernetaddr
         * Necessary for ODIPKT and logical boards
         */
       for (i=0;i<n;i++) {
         t=0;
	 for (ii=0;ii<=6;ii++) {
	   if (drivers[n].myeaddr[ii]==drivers[i].myeaddr[ii]) t++;
         }
	}
	/* put the interface into promiscuous mode */
	drivers[n].oldmode = set_mode(intno,drivers[n].handle,6);
	if (drivers[n].oldmode == -1) {
           printf("Error: driver %x cannot switch to promiscous mode\n",
                  intno);
           return (3);
        }

        printf("Driver at Int %d, handle %.4x, Adr ",
               intno, drivers[n].handle);
	printaddr(tmpstr, drivers[n].myeaddr);
	puts(tmpstr);
        if (driver_info(intno, drivers[n].handle, &version, &class,
	            &type, &number, &basic, drivers[n].name, NAMESIZE)) {
           printf(" - no info avail.");
        } else {
           if (drivers[n].name != NULL)
              printf("\n   '%s'",drivers[n].name);
           if (basic>=5) {
              printf(", 'high performance'");
              drivers[n].can_as=1;
           } else {
              drivers[n].can_as=0;
           }
        }
        putchar('\n');
        return 0;
}

/* release packet drivers */

void rel_drivers(void)
{
        int n;

        for (n=0;n<drvnum;n++) {
           if (drivers[n].intno==0) break;
	   printf("release driver %d int %d handle %.4x ... ",
                  n,drivers[n].intno,drivers[n].handle);
           set_mode(drivers[n].intno,drivers[n].handle,drivers[n].oldmode);
           if (release_type(drivers[n].intno, drivers[n].handle)) {
              printf("Error!!!\n");
              printf("It's highly recommended to REBOOT NOW!!!\n");
           } else
              printf("ok.\n");
	}
}

/* a packet with destination broadcast was received */
/* check if it is for us and send a proper answer */
/* return: 0 - packet processed, drop it */
/*         1 - not processed, handle as usual (broadcast) */

int process_arp(int intno, char far *buffer, unsigned int len)
{
     struct arp_packet *arp=(struct arp_packet *) buffer;

     if (!do_arp)    /* only if ARP enabled */
         return 1;

     gotoxy(1,20-drvnum);
     if (arp->ethtype!=0x0608) {    /* ARP ? */
	return 1; }                 /* no, finished */
#ifndef HIGHP
     if (arp->prot_type!=0x0008) {  /* IP ? */
#ifdef debug_ARP
	printf("ARP not IP\n");
#endif
	return 1; }                 /* no */
     if (arp->hw_type!=0x0100) {    /* ARP for ethernet ? */
	printf("ARP not Ethernet\n");
	return 1; }                 /* no */
#ifdef debug_ARP
     if (arp->hw_len!=6) {
	printf("ARP: wrong ethernet address length (this shouldn't happen)!\n");
	return 1; }
     if (arp->prot_len!=4) {
	printf("ARP: wrong IP address length (this shouldn't happen)!\n");
	return 1; }
#endif
     if (arp->opcode!=0x0100) {     /* ARP request ? (network-oder!) */
	printf("ARP but not a request\n");
	return 1; }                 /* no */
#endif /* Not defined HIGHP */

     if (memcmp(&arp->prot_dest, &my_ip, 4)) { /* our ip address ? */
#ifdef debug_ARP
	clreol();                   /* no */
	/* this is the output like tcpdump */
	printf("ARP who-is %s tell %s\n", sprint_IP(arp->prot_dest),
					  sprint_IP(arp->prot_send) );
#if 0
	printf("ARP who-is %u.%u.%u.%u tell %u.%u.%u.%u\n",
		arp->prot_dest.adr[0], arp->prot_dest.adr[1],
		arp->prot_dest.adr[2], arp->prot_dest.adr[3],
		arp->prot_send.adr[0], arp->prot_send.adr[1],
		arp->prot_send.adr[2], arp->prot_send.adr[3]);
#endif
#endif
	return 1; }
     memcpy(arp->destadr,arp->srcadr,6); /* prepare ARP reply */
     memcpy(arp->hw_dest,arp->srcadr,6);
     memcpy(arp->srcadr,Ethernetad,6);
     memcpy(arp->hw_send,Ethernetad,6);  /* this is what they want */
     arp->prot_dest=arp->prot_send;
     arp->prot_send=my_ip;
     arp->opcode=0x0200;       /* == ARP reply (network-order) */
#ifdef debug_ARP
     clreol();
     printf("sending ARP reply to %s\n",sprint_IP(arp->prot_dest) );
#if 0
     printf("sending ARP reply to %u.%u.%u.%u\n",arp->prot_dest.adr[0],
	     arp->prot_dest.adr[1],arp->prot_dest.adr[2],
	     arp->prot_dest.adr[3]);
#endif
#endif
     send_pkt(intno,buffer,len);
     arp_cnt++;
     return 0;
}

/* the main program */

int main(int argc, char * argv[])
{
	int n,problem=0,line=0;
	register int n1=0x60;
	int far *farscreen;
	char key=(char) 0;
#ifndef HIGHP
	char tmpstr[1500];
#endif
        
	clrscr();
	window(1,5,80,25); /* preserve some lines for the driver output */
	printf("yabri version %s\n\n",version);
	if (argc > 1) {
	  for (n=1; n<argc; n++) {
	    if (!strcmp(argv[n],"ip")) {
		unsigned int i1=0,i2=0,i3=0,i4=0;
		int erg;

		if (argc<n+2) {
		   printf("Missing arguments!\n");
		   print_help();
		   exit(1);
		}
		erg=sscanf(argv[n+1],"%u.%u.%u.%u",&i1,&i2,&i3,&i4);
		if (erg!=4) {
		   printf("Cannot translate '%s' to an IP address\n",argv[2]);
		   exit(1);
		}
		printf("IP address '%u.%u.%u.%u'",i1,i2,i3,i4);
		if ( (i1>254) || (i3>255) || (i3>255) || (i4>255) || (i1==0) ) {
		   printf(" invalid!\n");
		   exit(1);
		}
		if (i1==127) {
		   printf(" - loopback address not possible!\n");
		   exit(1);
		} else if (i1<127) {
		   printf(" Class-A net.");
		   if (i1!=10)
                      printf(" Are you sure?\n");
                   else
                      putchar('\n');
                   if ( (i2 & i3 & i4) == 255) {
		      printf("IP broadcast addresse not valid!\n");
                      exit(1); }
                } else if (i1>191) {
                   printf(" Class-C net.\n");
                   if ( i4 == 255) {
		      printf("IP broadcast addresse not valid!\n");
		      exit(1); }
                } else {
                   printf(" Class-B net.\n");
                   if ( (i3 & i4) == 255) {
		      printf("IP broadcast addresse not valid!\n");
		      exit(1); }
                }
                my_ip.adr[0]=i1;
                my_ip.adr[1]=i2;
                my_ip.adr[2]=i3;
		my_ip.adr[3]=i4;
                do_arp=1;
                n++;
            } else if (argv[n][0]=='/') {
		if (argv[n][1]=='h') {
                   help_start(argv[0]);
                   exit(0);
                } else if (argv[n][1]=='t') {
		   if (strchr(argv[n],'=')) {
		      key=strchr(argv[n],'=')[1];
                      Timer_on=1;
                   }
                } else if (argv[n][1]=='c') {
                   copyright_start(argv[0]);
		   exit(0);
                } else {
                   printf("Unknown option '%s'!\n",argv[n]);
                   exit(1);
                }
            } else {
                printf("Unknown parameter '%s'!\n",argv[n]);
                exit(1);
            } /* if 'ip' */
          }
        } else { /* indicates, that there is no IP address */
	    my_ip.adr[0]=0;
        } /* if argc > 1 */


        /* Search for graphic adapter at 0xb000 or 0xb800 */
	screen=0xb000;
        farscreen=(int far *)MK_FP(screen,0);
        for (n=0;n<5;n++) {
            *farscreen=n;
            *farscreen=-n*16;
	    if (*farscreen != -n*16) {
               screen=0xb800;
               break;
            }
	}
        for (n=0;n<5;n++)
            drivers[n].intno=0;
	/* Searching for packet drivers */
	drvnum=0;
	printf("\nTesting for packet drivers ...\n");
	for (n1 = 0x60; n1 <= 0x80; n1++) {
            if (test_for_pd(n1)) {
                switch (drvnum) {
                   case 0: line=newpkt(n1,receiver1); break;
		   case 1: line=newpkt(n1,receiver2); break;
                   case 2: line=newpkt(n1,receiver3); break;
                   case 3: line=newpkt(n1,receiver4); break;
                   case 4: line=newpkt(n1,receiver5); break;
                   default: printf("oops! Check for 6. driver.\n");
                            exit(1);
                }
                switch (line) {
                   case 1:
                   case 2:
                      printf("Error initializing packet drivers!\n");
		      rel_drivers();
                      exit(1);
                   case 3: problem=1; /* search further drivers */
                   default: ;
                }
		if (drivers[drvnum].intno!=0)
                   drvnum++;
                if (drvnum>=5) break;
            }
	}
	if (problem) exit(1); /* Problem with one driver? */

	if (drvnum==0) {
		printf("Error: No packet drivers found\n");
		exit(1);
	}

        if (drvnum==1) {
		printf("Warning: Only one packet driver found\n");
		printf("Start yabri? (y/n)?");
                if (getchar() != 'y') {
                   rel_drivers();
		   exit(1);
                }
	}
        for (n1=0;n1<6;n1++) {
            Ethernetad[n1]=drivers[0].myeaddr[n1];
        }
        if (my_ip.adr[0]==0) {   /* Setting IP address (?) */
            my_ip.adr[0]=10;     /* Class-A net 10.x.x.x (just to have one) */
            my_ip.adr[1]=Ethernetad[3];
            my_ip.adr[2]=Ethernetad[4];
            my_ip.adr[3]=Ethernetad[5];
        }

	printf("\nPress <h> for a short summary of commands\n");

        /* start Timer if the user wants this */
        if (Timer_on) get_TimerIRQ();

	/* and now the main loop */
	do {
	   for (n1=0;n1<PUFFERS;n1++) {
	     /* a full buffer has TxFlag != 0 */
	     if (recvbuf[n1].TxFlag==0) continue;
	     /* if pkt_len == 0 there is an error in the asm part! */
#ifndef HIGHP
	     if (recvbuf[n1].pkt_len==0) {
		fprintf(stderr,"Error: packet %d with length <null>\n",n1);
		problem = 1;
		/* we test for more errors before exiting */
	     }
#endif
	     /* TxFlag holds the number of the driver where the destination
		machine is located (normal packet) or which received
		the packet (broadcast packet) */
	     line = drivers[(recvbuf[n1].TxFlag & 0xff)-1].intno;
	     /* line == 0 should never occur */
#ifndef HIGHP
	     if (line==0) {
	       fprintf(stderr, "Error: invalid drivernumber %d\n",
		    recvbuf[n1].TxFlag);
	       problem=1;
	     }
	     if (problem) break;
#endif
	     /* TxFlag & 0x4000 -> packet with our ethernet address
		as destination address */
	     if (recvbuf[n1].TxFlag & 0x4000) {
		/* at this time we respond to ICMP Echo request and
		   UDP...*/
#if 0 /* War ifdef debug_ping */
		gotoxy(1,20-drvnum);
		printf("packet for us received from: ");
		printaddr( (char *) recvbuf[n1].buffer+6);
		putchar('\n');
#endif
		(void)process_pkt(line,recvbuf[n1].buffer,recvbuf[n1].pkt_len);
#if 0
		gotoxy(1,20-drvnum);
		printf("                            ");
#endif
	     /* TxFlag & 0x8000 -> packet is a broadcast packet */
	     } else if (recvbuf[n1].TxFlag & 0x8000) {
		/* we check if it's an ARP request to us */
		if (process_arp(line, recvbuf[n1].buffer, recvbuf[n1].pkt_len)) {
		   /* it's not for us so send it to all other lines */
		   for (n=0;n<drvnum;n++) {
		     /* send it to every line expect the one it came from
			Here line tells where the packet came from
		      */
		     if ((drivers[n].intno!=line) &&
			 (drivers[n].intno!=0)) {
			 send_pkt(drivers[n].intno,recvbuf[n1].buffer,recvbuf[n1].pkt_len);
		     }
		   }
		} /* if process_arp */
	     } else {
		/* not broadcast nor to us so send it to specified line
		   here line tells where to send the packet */
		send_pkt(line,recvbuf[n1].buffer,recvbuf[n1].pkt_len);
	     }

	     /* clear the buffer - order is important! */
	     recvbuf[n1].TxFlag = 0;
	     recvbuf[n1].pkt_len = 0;
	     recvbuf[n1].drvr = 0;
	   } /* for */
	   /* was there an error? (fatal is set by the asm part */


#ifndef HIGHP

	   if (fatal || problem) {
	      fprintf(stderr, "Fatal Error: fatal:%d problem:%d\n",fatal,problem);
	      break;
	   }
	   /* handle user input */

	   if (kbhit()) {
	     key=getch();
	     switch (key) {
	     case '\033':  /* esc */
	     case 'q':
	     case ' ':
	       problem=1;  /* indicate stop */
	       break;
	     case 't':     /* timer-irq on/off */
	       if (!Timer_on) {
		  get_TimerIRQ();
		  Timer_on=1;
	       } else {
		  release_TimerIRQ();
		  Timer_on=0;
	       }
	       break;
	     case 'p':     /* print address table */
	       clrscr();
	       print_drivers(tmpstr); puts(tmpstr);
	       print_table(tmpstr); puts(tmpstr);
	       break;
	     case 'f':     /* flush address table */
	       flush_table();
	       break;
	     case 'i':     /* print info / help */
	       clrscr();
	       print_info(tmpstr); puts(tmpstr);
	       break;
	     case 'h':
	       clrscr();
	       print_help();
	       break;
	     case 's':     /* statistics */
	       clrscr();
	       print_drivers(tmpstr); puts(tmpstr);
	       print_stat(tmpstr); puts(tmpstr);
	       break;
	     case 'R':
	       if (!repeater) {
		  clrscr();
		  printf("switching to repeater mode\n");
		  repeater=1;
	       } else {
		  printf("switching from repeater to normal mode \n");
		  print_stat(tmpstr); puts(tmpstr);
		  repeater=0; }
	       break;
	     case 'a':
	       if (do_arp) {
		  printf("ARP reply off\n");
		  do_arp=0;
	       } else {
		  printf("ARP reply on\n");
		  do_arp=1;
	       }
	       break;
	     default:
	       sound(5000);  /* beep a little bit */
	       delay(5);
	       nosound();
	     key=(char)0; /* paranoia setting */
	     } /* case */
	   } /* if kbhit */
#endif /* HIGHP */
	} while(!problem);
	/* release timer and drivers on exit.
	   if an error occurs it is possible that the timer / drivers
	   still call our upcalls - which are no longer there! The only
	   way to get around this is to reboot */
        if (Timer_on) release_TimerIRQ();
	rel_drivers();
	return 0;
}



